"""Python utilities for working with nucleotide sequence strings.

Variety of utilities for converting, searching, and doing 
calculations on nucleotide sequences.

"""

from collections.abc import Generator, Sequence
from difflib import SequenceMatcher
import re

from .utils import (sequences, 
                    _preserve_case, 
                    _preserve_circular,
                    _normalize_case)

seqs = sequences


@_preserve_circular
def reverse(x: str) -> str:

    """Reverse a sequence.

    Parameters
    ----------
    x : str
        Sequence to convert.

    Returns
    -------
    str
        Converted sequence.

    """

    return x[::-1]


@_preserve_circular
@_preserve_case
@_normalize_case(nargs=1)
def complement(x: str) -> str:

    """Complement (but don't reverse) a sequence.

    Parameters
    ----------
    x : str
        Sequence to convert.

    Returns
    -------
    str
        Converted sequence.

    """

    return x.translate(seqs.complementer)


def reverse_complement(x: str) -> str:

    """Reverse complement a sequence.

    Parameters
    ----------
    x : str
        Sequence to convert.

    Returns
    -------
    str
        Converted sequence.

    Examples
    --------
    >>> reverse_complement('ATCG')
    'CGAT'

    """

    return complement(reverse(x))


@_preserve_circular
@_preserve_case
@_normalize_case(nargs=1)
def to_rna(x: str) -> str:

    """Convert nucleotides to RNA.

    Parameters
    ----------
    x : str
        Sequence to convert.

    Returns
    -------
    str
        Converted sequence.

    Examples
    --------
    >>> to_rna('ATCG')
    'AUCG'

    """

    return x.replace('T', 'U')


@_preserve_circular
@_preserve_case
@_normalize_case(nargs=1)
def to_dna(x: str) -> str:

    """Convert nucleotides to DNA.

    Parameters
    ----------
    x : str
        Sequence to convert.

    Returns
    -------
    str
        Converted sequence.

    Examples
    --------
    >>> to_dna('AUCG')
    'ATCG'

    """

    return x.replace('U', 'T')


@_normalize_case(nargs=2)
def find_iupac(query: str, 
               sequence: str) -> Generator[Sequence[int], str]:
    
    """Find occurrences of a query in a larger sequence.

    IUPAC codes in the query will be interpreted as ambiguities:

    A: A
    C: C
    G: G
    T: T
    U: U
    N: .
    R: "[AG]"
    Y: "[TUC]"
    W: "[ATU]"
    S: "[CG]"
    V: "[ACG]"
    B: "[TUGC]"

    Parameters
    ----------
    query : str
        Sequence to search for. Accepts IUPAC codes: N, R, Y, S, W, V, B.
    sequence : str 
        Sequence to search within.
    
    Yields
    ------
    Generator
        Generator of tuples containing the match indices and matched sequence.
    indices : tuple
        Start and stop indices of the match
    sequence : str
        matched sequence

    Examples
    --------
    >>> for (start_idx, end_idx), match in find_iupac('ARY', 'AATAGCAGTGTGAAC'):
    ...     print(f"Found ARY at {start_idx}:{end_idx}: {match}")
    ... 
    Found ARY at 0:3: AAT
    Found ARY at 3:6: AGC
    Found ARY at 6:9: AGT
    Found ARY at 12:15: AAC

    """
    
    query = query.translate(seqs.base2regex)
    query = re.compile(query) 
    
    for match in query.finditer(sequence):

        yield match.span(), match.group()


@_normalize_case(nargs=1)
def which_re_sites(x: str) -> Sequence[str]:

    """List Type IIS restriction sites in sequence.

    Currently only searches for the most commonly used
    Type IIS restriction sites for Golden Gate Cloning:

    BbsI: GAAGAC
    BsmBI: CGTCTC
    BtgZI: GCGATG
    PaqCI: CACCTGC
    SapI: GCTCTTC
    BsaI: GGTCTC

    Parameters
    ----------
    x : str
        Sequence to check.

    Returns
    -------
    tuple
        List of Type IIS restriction sites in x

    Examples
    --------
    >>> which_re_sites('AAAGAAG')
    ()
    >>> which_re_sites('AAAGAAGAC')
    ('BbsI',)
    >>> which_re_sites('AAAGAAGACACCTGC')
    ('BbsI', 'PaqCI')
    
    """

    fwd = [enz for enz, site in seqs.re_sites.items() 
           if (site in x) or 
           (reverse_complement(site) in x)]

    return tuple(fwd)


@_normalize_case(nargs=1)
def count_re_sites(x: str) -> bool:

    """Count Type IIS restriction sites in sequence.

    Currently only searches for the most commonly used
    Type IIS restriction sites for Golden Gate Cloning:

    BbsI: GAAGAC
    BsmBI: CGTCTC
    BtgZI: GCGATG
    PaqCI: CACCTGC
    SapI: GCTCTTC
    BsaI: GGTCTC

    Parameters
    ----------
    x : str
        Sequence to check.

    Returns
    -------
    int
        Number of Type IIS restriction sites in x.

    Examples
    --------
    >>> count_re_sites('AAAGAAG')
    0
    >>> count_re_sites('AAAGAAGAC')
    1
    >>> count_re_sites('AAAGAAGACACCTGC')
    2
    
    """

    return len(which_re_sites(x))
    

@_normalize_case(nargs=2)
def _x_content(x: str, y: str) -> float:

    try:
        return sum(letter in y for letter in x) / len(x)
    except ZeroDivisionError:
        return 0. 
    

def gc_content(x: str) -> float:

    """Calculate proportional GC content.

    Recognises IUPAC codes.

    Parameters
    ----------
    x : str
        Sequence.

    Returns
    -------
    float
        GC content.

    Examples
    --------
    >>> gc_content('AGGG')
    0.75

    """

    return _x_content(x, 'GCS')


def purine_content(x: str) -> float:

    """Calculate proportional purine content.

    Recognises IUPAC codes.

    Parameters
    ----------
    x : str
        Sequence.

    Returns
    -------
    float
        Purine content.

    Examples
    --------
    >>> purine_content('AUGGR')
    0.8

    """

    return _x_content(x, 'GAR')


def pyrimidine_content(x: str) -> float:

    """Calculate proportional pyrimidine content.

    Recognises IUPAC codes.

    Parameters
    ----------
    x : str
        Sequence.

    Returns
    -------
    float
        Pyrimidine content.
    
    Examples
    --------
    >>> pyrimidine_content('AUGGG')
    0.2

    """

    return _x_content(x, 'CUTY')
    


"""Classes relating to circular strings."""


class Circular(str):

    """A string-like object which can be circularly sliced.

    Useful for sequences that represent a bacterial genome 
    or a plasmid.

    Currently this only works if the length of the slice is 
    shorter than the sequence's total length.

    Methods
    -------
    __getitem__
        Slice string as if it were a circle.

    Examples
    --------
    >>> Circular('ATCG')[:3]
    'ATC'
    >>> Circular('ATCG')[-1:3]
    'GATC'
    >>> from streq import reverse_complement
    >>> reverse_complement(Circular('ATCg'))
    'cGAT'

    """

    def __getitem__(self, __key: slice) -> str:

        """Slice circularly.
        
        Parameters
        ----------
        __key : slice
            Python slice

        Returns
        -------
        str
            Slice from Circular object. Ordinary string.
        
        """

        if isinstance(__key, int):

            return super().__getitem__(__key)
        
        elif isinstance(__key, slice):

            start = __key.start or 0
            stop = (len(self) if __key.stop is None 
                    else __key.stop) 

            return (super().__getitem__(slice(start, None)) + self)[:(stop - start)][::__key.step]
        


from difflib import SequenceMatcher

from .seqtools import reverse_complement
from .utils import _normalize_case

@_normalize_case(nargs=2)
def levenshtein(x: str, y: str) -> int:

	"""Calculate the Levenshtein distance between two sequences.
    
    The Levenshtein distance is the number of insertions,
    deletions, and mutations required to make two 
    sequences match.

    Parameters
    ----------
    x : str
        Sequence.
    y : str, optional
        Second sequence for correlation with x.

    Returns
    -------
    int
        Levenshtein distance.

    Examples
    --------
    >>> levenshtein('AAATTT', 'AAATTT')
    0
    >>> levenshtein('AAATTT', 'ACTTT')
    2
    >>> levenshtein('AAATTT', 'AACTTT')
    1
    >>> levenshtein('AAAG', 'TCGA')
    4

    """
    
	x_len, y_len = map(len, (x, y))

	matrix = {i: {j: 0 for j in range(y_len + 1)} 
			  for i in range(x_len + 1)}
	
	for i in range(x_len + 1):

		matrix[i][0] = i
	
	for j in range(y_len + 1):

		matrix[0][j] = j

	for i in range(1, x_len + 1):
		
		for j in range(1, y_len + 1):
			
			if x[i - 1] == y[j - 1]:
				
				matrix[i][j] = matrix[i - 1][j - 1]
			
			else:
				
				minimum = min(matrix[i - 1][j], 
							  matrix[i][j - 1], 
							  matrix[i - 1][j - 1])
				matrix[i][j] = minimum + 1

	return matrix[x_len][y_len]


@_normalize_case(nargs=2)
def hamming(x: str, y: str, 
	        wobble: bool = False) -> int:
    
    """Calculate the Hamming distance between two sequences.
    
    The Hamming distance is the number of mismatches.

    Note: the `wobble` parameter is not yet implemented.

    Parameters
    ----------
    x : str
        Sequence.
    y : str, optional
        Second sequence for correlation with x.
    wobble : bool, optional
        Whether to calulate correlations taking into account G.U wobble.
        Not yet implemented.

    Returns
    -------
    int
        Hamming distance.

    Examples
    --------
    >>> hamming('AAA', 'ATA')
    1
    >>> hamming('AAA', 'ATT')
    2
    >>> hamming('AAA', 'TTT')
    3

    """

    return sum(a != b for a, b in zip(x, y))


@_normalize_case(nargs=2)
def ratcliff_obershelp(x: str, 
					   y: str) -> int:
    
	"""Calculate the Ratcliff-Obershelp distance between two sequences.
    
    The Ratcliff-Obershelp distance is the number of grouped
	insertions, deletions, and mutations required to make two 
    sequences match.

    Parameters
    ----------
    x : str
        Sequence.
    y : str, optional
        Second sequence for correlation with x.

    Returns
    -------
    int
        Ratcliff-Obershelp distance.

    Examples
    --------
    >>> ratcliff_obershelp('AAATTT', 'AAATTT')
    0
    >>> ratcliff_obershelp('AAATTT', 'ACTTT')
    1
    >>> ratcliff_obershelp('AAATTT', 'AACTTT')
    1
    >>> ratcliff_obershelp('AAAG', 'TCGA')
    2
    
    """
    
	sm = SequenceMatcher(a=x, b=y, 
                         autojunk=False).get_opcodes()
    
	return sum(code[0] != 'equal' for code in sm)


@_normalize_case(nargs=2)
def correlation(x: str, 
                y: str = '',
                wobble: bool = False) -> float:
    
    """Get autocorrelation of a single sequence or 
    correlation between two sequences.

    If a single sequence is provided, it's autocorrelation 
    is calculated, which might be an indicator of secondary structure.

    Note: the `wobble` parameter is not yet implemented.

    Parameters
    ----------
    x : str
        Sequence.
    y : str, optional
        Second sequence for correlation with x.
    wobble : bool, optional
        Whether to calulate correlations taking into account G.U wobble.
        Not yet implemented.

    Returns
    -------
    float
        Correlation.

    Examples
    --------
    >>> correlation('AACC')
    0.0
    >>> correlation('AAATTT')
    2.3
    >>> correlation('AAATTCT')
    1.3047619047619046
    >>> correlation('AAACTTT')
    1.9238095238095236
    >>> correlation('AAA', 'TTT')
    0.0
    >>> correlation('AAA', 'AAA')
    3.0

    """
    
    x = x
    y = y or reverse_complement(x)

    max_len = min(len(x), len(y))
    
    return sum(((max_len - n) - hamming(x[n:], y, wobble)) / (max_len - n)
               for n in range(len(x)))


"""Miscellaneous utilities used in streq."""

from collections import namedtuple
from collections.abc import Callable, Sequence
from functools import wraps
import os

import yaml

from .circular import Circular

_data_path = os.path.join(os.path.dirname(__file__), 
                          'sequences.yml')

with open(_data_path, 'r') as f:
    _sequences = yaml.safe_load(f)

sequence_dict = dict(
    complementer = str.maketrans(_sequences['complement']),
    re_sites = _sequences['type_iis_re'],
    DNA = _sequences['alphabet']['DNA'],
    RNA = _sequences['alphabet']['RNA'],
    base2regex = str.maketrans(_sequences['base2regexp']),
    PAMs = _sequences['PAMs']
)

SequenceCollection = namedtuple('SequenceCollection',
                                sequence_dict)

sequences = SequenceCollection(**sequence_dict)


def _make_lower(x: str, 
                lower: Sequence[bool]):
    
    return ''.join(str.lower(letter) if low 
                   else letter 
                   for letter, low in zip(x, lower))


def _normalize_case(nargs: int = 1) -> Callable[[Callable], Callable]:
    
    def decor(f: Callable[[str, str], 
                               str]) -> Callable[[str, str], 
                                                 str]:
        
        @wraps(f)
        def normalized(*args, **kwargs):

            args = (x.casefold().upper() 
                    for i, x in enumerate(args) if i < nargs)
            
            return f(*args, **kwargs)
        
        return normalized
    
    return decor


def _preserve_case(f: Callable[[str], 
                               str]) -> Callable[[str], 
                                                 str]:

    @wraps(f)
    def preserved(x: str, 
                  *args, **kwargs):
        
        is_lower = (letter.islower() for letter in x)

        x = f(x.casefold().upper(), *args, **kwargs)
        
        return _make_lower(x, is_lower)
    
    preserved.__doc__ = (f.__doc__ or '') + (
    """
    Note: Preserves case.

    """
    )
    
    return preserved


def _preserve_circular(f: Callable[[str], 
                                    str]) -> Callable[[str], 
                                                  str]:
    
    @wraps(f)
    def preserved(x: str, 
                  *args, **kwargs):

        is_circular = isinstance(x, Circular)

        x = f(x, *args, **kwargs)

        if is_circular:

            x = Circular(x)

        return x
    
    preserved.__doc__ = (f.__doc__ or '') + (
    """
    Note: Preserves circularity.

    """
    )
    
    return preserved